/* $%BEGINLICENSE%$
 Copyright (c) 2012, Oracle and/or its affiliates. All rights reserved.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License as
 published by the Free Software Foundation; version 2 of the
 License.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 02110-1301  USA

 $%ENDLICENSE%$ */
 

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <glib.h>
#include "chassis-plugin-stats.h"

chassis_plugin_local_stats_t *plugin_global_stats = NULL;
GList *ts_read_query_stats_list = NULL;
network_backends_t *backends_stats = NULL;

/**
 * chassis_plugin_stats_new:
 * @stats: pointer to a #chassis_plugin_local_stats_t
 *
 * init the local plugin stats chassis_plugin_local_stats
 *
 * Returns: a pointer to the newly created #chassis_plugin_local_stats_t
 */
chassis_plugin_local_stats_t *chassis_plugin_stats_new(chassis_plugin_local_stats_t *stats) {
	if (stats != NULL) return stats;
	
	stats = g_new0(chassis_plugin_local_stats_t, 1);
	g_debug("%s: created new plugin specific chassis stats at %p", G_STRLOC, (void*)stats);

	/* initialize plugin_global_stats as well */
	plugin_global_stats = stats;
	
	return stats;
}

/**
 * chassis_plugin_stats_free:
 * @stats: pointer to a #chassis_plugin_local_stats_t
 *
 * free the local plugin stats chassis_plugin_local_stats
 */
void chassis_plugin_stats_free(chassis_plugin_local_stats_t *stats) {
	if (!stats) return;
	
	g_free(stats);
}

/**
 * chassis_plugin_stats_get:
 * @stats: pointer to a #chassis_plugin_local_stats_t
 *
 * get a hash(str, int) of all local plugin statistic values
 *
 * Returns: (element-type gchar int): a hash
 */
GHashTable* chassis_plugin_stats_get(chassis_plugin_local_stats_t *stats){
	GHashTable *stats_hash;
	
	if (stats == NULL) return NULL;
	
	/* NOTE: the keys are strdup'ed, the values are simply integers */
	stats_hash = g_hash_table_new_full(g_str_hash, g_str_equal, g_free, g_free);

#define STR(x) #x
#define N(x) g_strdup(x)
#define ADD_STAT(x) g_hash_table_insert(stats_hash, N( STR(x)), GUINT_TO_POINTER(g_atomic_int_get(&(stats->x))))
#define ADD_UPTIME_STAT(x) g_hash_table_insert(stats_hash, N( STR(x)), chassis_plugin_stats_uptime_get_pointer(stats))
#define ADD_ALLOC_STAT(x) ADD_STAT(x ## _alloc); ADD_STAT(x ## _free);
	
	ADD_UPTIME_STAT(plugin_init_timestamp);
	ADD_STAT(read_query_ts_avg);
	ADD_STAT(read_query_ts_min);
	ADD_STAT(read_query_ts_max);
	ADD_STAT(connected_clients);
	ADD_STAT(connected_clients_max);
	
#undef N
#undef STR
#undef ADD_STAT
#undef ADD_ALLOC_STAT
	
	return stats_hash;
}

/**
 * chassis_plugin_stats_uptime_set:
 * @user_data: pointer to the #chassis_plugin_local_stats_t
 * @timestamp: time_t containing the timestamp to set
 *
 * sets a time_t value of timestamp on the plugin local #chassis_plugin_local_stats_t
 */
void chassis_plugin_stats_uptime_set(chassis_plugin_local_stats_t *user_data, time_t timestamp) {
	/* let's initialize the static mutex */
	static GStaticMutex mutex = G_STATIC_MUTEX_INIT;

	g_static_mutex_lock(&mutex);
	user_data->plugin_init_timestamp = timestamp;
	g_static_mutex_unlock(&mutex);
}

/**
 * chassis_plugin_stats_uptime_get:
 * @user_data: pointer to the #chassis_plugin_local_stats_t
 *
 * gets the #plugin_init_timestamp value of the plugin local #chassis_plugin_local_stats_t
 *
 * Returns: a time_t value containing the requested local #plugin_init_timestamp
 */
time_t chassis_plugin_stats_uptime_get(chassis_plugin_local_stats_t *user_data) {
	static GStaticMutex mutex = G_STATIC_MUTEX_INIT;

	g_static_mutex_lock(&mutex);
	time_t plugin_init_timestamp = user_data->plugin_init_timestamp;
	g_static_mutex_unlock(&mutex);

	return plugin_init_timestamp;
}

/**
 * chassis_plugin_stats_uptime_get_pointer:
 * @user_data: pointer to the #chassis_plugin_local_stats_t
 *
 * allocates memory and gets a pointer to the #plugin_init_timestamp value of the plugin local #chassis_plugin_local_stats_t
 *
 * Returns: a pointer to the newly allocated time_t containing the requested local #plugin_init_timestamp
 */
time_t *chassis_plugin_stats_uptime_get_pointer(chassis_plugin_local_stats_t *user_data) {
	static GStaticMutex mutex = G_STATIC_MUTEX_INIT;

	time_t *plugin_init_timestamp = g_new0(time_t, 1);

	g_static_mutex_lock(&mutex);
	plugin_init_timestamp = (time_t*)user_data->plugin_init_timestamp;
	g_static_mutex_unlock(&mutex);

	return plugin_init_timestamp;
}

/**
 * chassis_plugin_stats_read_query_stats_set:
 * @time_t: time_t containing the timestamp of the obtainance of the #ts_diff
 * @ts_diff: guint value of the difference between #ts_read_query_result_last and #ts_read_query_result_first
 *
 * sets a time_t value of timestamp and a guint value of the difference between
 * #ts_read_query_result_last and ts_read_query_result_first on the global list ts_read_query_stats_list
 */
void chassis_plugin_stats_read_query_stats_set(time_t timestamp, guint ts_diff) {
	static GStaticMutex mutex = G_STATIC_MUTEX_INIT;

	ts_read_query_stats *ts_read_query_stats_aux = g_new0(ts_read_query_stats, 1);
	ts_read_query_stats_aux->get_timestamp = timestamp;
	ts_read_query_stats_aux->ts_diff = ts_diff;

	g_static_mutex_lock(&mutex);
	ts_read_query_stats_list =
		 g_list_append(ts_read_query_stats_list, ts_read_query_stats_aux);
	g_static_mutex_unlock(&mutex);
}

/**
 * chassis_plugin_stats_read_query_stats_get_avg:
 *
 * calculates the average query latency value
 *
 * Returns: a guint value of the calculated average of query latency
 */
guint chassis_plugin_stats_read_query_stats_get_avg(void) {
	guint count=0, ts_total=0, ts_avg;
	GList *iterator = NULL;

	for(iterator = ts_read_query_stats_list; iterator; iterator = iterator->next) {
		ts_total += ((ts_read_query_stats*)iterator->data)->ts_diff;
		count += 1;
	}

	ts_avg = (guint)(ts_total/count); 

	return ts_avg;
}

gint compare_tsdiff(ts_read_query_stats *a, ts_read_query_stats *b) {

	if(a->ts_diff > b->ts_diff)
		return 1;
	
	if(a->ts_diff < b->ts_diff)
		return -1;

	else
		return 0;
}

/**
 * chassis_plugin_stats_read_query_stats_get_min:
 *
 * calculates the minimum query latency value
 *
 * Returns: a guint value of the minimum value of query latency
 */
guint chassis_plugin_stats_read_query_stats_get_min(void) {

	ts_read_query_stats_list = g_list_sort(ts_read_query_stats_list, (GCompareFunc)compare_tsdiff);

	return (((ts_read_query_stats*)ts_read_query_stats_list->data)->ts_diff);
}

 /**
 * chassis_plugin_stats_read_query_stats_get_max:
 *
 * calculates the maximum query latency value
 *
 * Returns: a guint value of the maximum value of query latency
 */
 guint chassis_plugin_stats_read_query_stats_get_max(void) {

 	ts_read_query_stats_list = g_list_sort(ts_read_query_stats_list, (GCompareFunc)compare_tsdiff);

 	ts_read_query_stats_list = g_list_last(ts_read_query_stats_list);

	return (((ts_read_query_stats*)ts_read_query_stats_list->data)->ts_diff);
}
